from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np


class Palette:
    """Colour palette methods."""

    pixels: np.ndarray

    def palette(self, **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        """Compute the colour distribution.

        Args:
            **kwargs: passed to `numpy.histogram`.

        Returns:
            Histogram edges and counts.
        """
        bins = kwargs.pop("bins", 256)
        values = []
        bin_edges = []
        for i in range(self.pixels.shape[1]):
            freqs, edges = np.histogram(self.pixels[:, i], bins=bins, **kwargs)
            bin_edges.append(edges)
            values.append(freqs)
        values = np.vstack(values).T
        bin_edges = np.vstack(bin_edges).T
        return bin_edges, values

    def cdfs(self):
        """Compute the cumulative distribution functions of the colours ditributions.

        Returns:
            Histogram edges and counts.
        """
        bins, frequencies = self.palette()
        return bins, self._cdfs(frequencies)

    @staticmethod
    def _cdfs(frequencies: np.ndarray) -> np.ndarray:
        cdfs = np.cumsum(frequencies, axis=0, dtype=float)
        cdfs = np.insert(cdfs, 0, 0, axis=0)
        cdfs /= cdfs[-1]
        return cdfs

    def plot(self, log: bool = False) -> Tuple[plt.Figure, plt.Axes]:
        """Plot the colour distribution.

        Args:
            log: Plot y axis in log scale.

        Returns:
            Plot figure and axes.
        """

        bin_edges, values = self.palette()
        fig, axes = plt.subplots(3, figsize=(12, 6))
        for i, ax in enumerate(axes):
            ax.bar(
                bin_edges[:-1, i],
                values[:, i],
                width=np.diff(bin_edges[:, i]),
                align="edge",
            )
            if log:
                ax.set_yscale("log")
            ax.set_title(f"Channel {i+1}")
        fig.tight_layout()
        return fig, axes



import logging
from typing import Iterator, List, Tuple

import numpy as np
from PIL import Image

from .master import Master


class Grid:
    def __init__(
        self,
        master: Master,
        mosaic_shape: Tuple[int, int],
        tile_shape: Tuple[int, int],
    ) -> None:
        """Mosaic tile grid.

        Args:
            master: Master image.
            mosaic_shape: mosaic image shape.
            tile_shape: size of the tiles.
        """
        self.master = master
        self.mosaic_shape = mosaic_shape
        self.tile_shape = tile_shape
        self.thresholds = []
        self.origin = (
            self.master.array.shape[0] % self.tile_shape[0] // 2,
            self.master.array.shape[1] % self.tile_shape[1] // 2,
        )
        self._log = logging.getLogger(__name__)
        self._slices = None
        self._arrays = None

    @property
    def arrays(self) -> List[np.ndarray]:
        """List of arrays containing the pixel values of all the slices."""
        if self._arrays is None:
            self._log.debug("Computing arrays.")
            self._arrays = [
                self.master.array[slices[0], slices[1]] for slices in self.slices
            ]
        return self._arrays

    @property
    def slices(self) -> List[Tuple[slice, slice]]:
        """Mosaic grid slices."""
        if self._slices is None:
            self._log.debug("Computing slices.")
            self._slices = list(self._compute_slices())
        return self._slices

    def _compute_slices(self) -> Iterator[Tuple[slice, slice]]:
        for x in range(
            self.origin[1], self.mosaic_shape[1] - self.origin[1], self.tile_shape[1]
        ):
            for y in range(
                self.origin[0],
                self.mosaic_shape[0] - self.origin[0],
                self.tile_shape[0],
            ):
                yield (
                    slice(y, y + self.tile_shape[0]),
                    slice(x, x + self.tile_shape[1]),
                )

    def remove_origin(self, slices: Tuple[slice, slice]) -> Tuple[slice, slice]:
        """Remove the origin of the mosaic from a slice tuple."""
        return (
            slice(slices[0].start - self.origin[0], slices[0].stop - self.origin[0]),
            slice(slices[1].start - self.origin[1], slices[1].stop - self.origin[1]),
        )

    @staticmethod
    def _subdivide(
        slices: Tuple[slice, slice]
    ) -> Tuple[
        Tuple[slice, slice],
        Tuple[slice, slice],
        Tuple[slice, slice],
        Tuple[slice, slice],
    ]:
        """Create four tiles from the four quadrants of the input tile.

        Examples:
            Subdivide a grid element.

            >>> mosaic.grid._subdivide((slice(0, 30, None), slice(0, 45, None)))
            ((slice(0, 15, None), slice(0, 22, None)),
             (slice(0, 15, None), slice(22, 45, None)),
             (slice(15, 30, None), slice(0, 22, None)),
             (slice(15, 30, None), slice(22, 45, None)))
        """
        height = slices[0].stop - slices[0].start
        width = slices[1].stop - slices[1].start
        tile_dims = [(s.stop - s.start) // 2 for s in slices]
        return (
            (
                slice(slices[0].start, slices[0].start + tile_dims[0]),
                slice(slices[1].start, slices[1].start + tile_dims[1]),
            ),
            (
                slice(slices[0].start, slices[0].start + tile_dims[0]),
                slice(
                    slices[1].start + tile_dims[1],
                    slices[1].start + 2 * tile_dims[1] + width % 2,
                ),
            ),
            (
                slice(
                    slices[0].start + tile_dims[0],
                    slices[0].start + 2 * tile_dims[0] + height % 2,
                ),
                slice(slices[1].start, slices[1].start + tile_dims[1]),
            ),
            (
                slice(
                    slices[0].start + tile_dims[0],
                    slices[0].start + 2 * tile_dims[0] + height % 2,
                ),
                slice(
                    slices[1].start + tile_dims[1],
                    slices[1].start + 2 * tile_dims[1] + width % 2,
                ),
            ),
        )

    def _yield_subdivide(self, threshold: float) -> Iterator[Tuple[slice, slice]]:
        n_divisions = 0
        for i, slices in enumerate(self.slices):
            pixels = self.master.array[slices[0], slices[1]].reshape(-1, 3)
            contrast = np.mean(np.std(pixels / 255, axis=0))
            self._log.debug("Contrast slice %i: %f", i, contrast)
            if contrast > threshold:
                self._log.debug("Dividing slice %i", i)
                yield from self._subdivide(slices)
                n_divisions += 1
            else:
                yield slices
        self._log.info("Subdivided %i tiles.", n_divisions)

    def subdivide(self, threshold: float) -> None:
        """Subdivide grid based on contrast.

        Note:
            Modifies the grid in place.

        Args:
            threshold: contrast threshold at which to divide the slice into 4
                smaller slices. Between 0 and 1.
        """
        self._log.info("Subdividing with threshold %f", threshold)
        self._slices = list(self._yield_subdivide(threshold))
        # clear the cached arrays
        self._arrays = None
        self.thresholds.append(threshold)

    def plot(self, colour: Tuple[int, int, int] = (255, 255, 255)) -> Image.Image:
        """Draw the tile edges on a copy of master image image.

        Args:
            colour: Colour of tile edges.

        Returns:
            Master image with the tile grid overlayed ontop.
        """
        annotated_image = self.master.array.copy()
        for y, x in self.slices:
            annotated_image[y, x.start] = colour
            annotated_image[y, x.stop - 1] = colour
            annotated_image[y.start, x] = colour
            annotated_image[y.stop - 1, x] = colour
        return Image.fromarray(annotated_image)

    def __len__(self) -> int:
        return len(self.slices)

    def __repr__(self) -> str:
        return f"""{self.__class__.__module__}.{self.__class__.__name__} at {hex(id(self))}:
    origin: {self.origin}
    len slices: {len(self.slices)}
    thresholds: {self.thresholds}"""



import logging
from pathlib import Path
from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
from PIL import Image
from tqdm.auto import tqdm

from .palette import Palette
from .utils import open_img_file


class Pool(Palette):
    @classmethod
    def from_dir(
        cls,
        tile_dir: Union[str, Path],
        *args,
        crop_ratio: Optional[float] = None,
        tile_size: Optional[Tuple[int, int]] = None,
        convert: Optional[str] = None,
        **kwargs,
    ) -> "Pool":
        """Create a Pool instance from the images in a directory.

        Args:
            tile_dir: path to directory containing the images.
            crop_ratio: width to height ratio to crop the master image to. 1 results in a square image.
            tile_size: resize the image to the provided size, width followed by height.
            convert: convert the image to the provided mode. See PIL Modes.
        """
        if not isinstance(tile_dir, Path):
            tile_dir = Path(tile_dir)
        if not tile_dir.is_dir():
            raise ValueError(f"'{tile_dir}' is not a directory.")
        arrays = cls._load_files(
            list(tile_dir.glob("*")),
            crop_ratio=crop_ratio,
            size=tile_size,
            convert=convert,
        )
        return cls(arrays, *args, **kwargs)

    @classmethod
    def from_files(
        cls,
        files: Sequence[Path],
        *args,
        crop_ratio: Optional[float] = None,
        tile_size: Optional[Tuple[int, int]] = None,
        convert: Optional[str] = None,
        **kwargs,
    ) -> "Pool":
        """Create a Pool instance from a list of images.

        Args:
            files: list of paths to the tile images.
            crop_ratio: width to height ratio to crop the master image to. 1 results in a square image.
            tile_size: resize the image to the provided size, width followed by height.
            convert: convert the image to the provided mode. See PIL Modes.
        """
        arrays = cls._load_files(
            files, crop_ratio=crop_ratio, size=tile_size, convert=convert
        )
        return cls(arrays, *args, **kwargs)

    def __init__(
        self,
        arrays: Sequence[np.ndarray],
    ) -> None:
        """A Pool of images tiles.

        Args:
            arrays: list of arrays containing the image pixel values. Should
                contain 3 colour channels.
        """
        self.arrays = np.stack(arrays)
        self._log = logging.getLogger(__name__)
        self._log.info("Number of tiles: %s", len(self.arrays))

    @property
    def tiles(self) -> "PoolTiles":
        """Access the Pool's tile images.

        Examples:
            Show the first image in the pool.

            >>> pool.tiles[0].show()
        """
        return PoolTiles(self.arrays)

    @property
    def pixels(self) -> np.ndarray:
        """Array containing the 3-channel pixel values of all the images in the Pool."""
        # if self._colors is None:
        #     self._log.debug("Computing colors.")
        #     self._colors = self._flatten_arrays(self.arrays)
        return np.vstack([array.reshape(-1, array.shape[-1]) for array in self.arrays])

    @staticmethod
    def _load_files(files: Sequence[Path], **kwargs) -> List[np.ndarray]:
        arrays = []
        for tile in tqdm(files, desc="Loading tiles"):
            img = open_img_file(tile, **kwargs)
            array = np.asarray(img)
            # make sure the arrays have 3 channels even in black and white
            if array.ndim == 2:
                array = np.stack([array] * 3, -1)
            arrays.append(array)
        return arrays

    def __len__(self) -> int:
        return len(self.arrays)

    def __repr__(self) -> str:
        return f"""{self.__class__.__module__}.{self.__class__.__name__} at {hex(id(self))}:
    len: {self.__len__()}"""


class PoolTiles:
    """Helper interface to access of PIL.Image instances of the tiles."""

    def __init__(self, arrays: np.ndarray) -> None:
        self._arrays = arrays

    def __getitem__(self, index) -> Image.Image:
        return Image.fromarray(self._arrays[index].round(0).astype("uint8"), mode="RGB")



import logging
from functools import partial
from multiprocessing.pool import Pool as MpPool
from typing import Optional, Tuple, Union

import numpy as np
from PIL import Image
from tqdm.auto import tqdm

from .grid import Grid
from .master import Master
from .metrics import METRICS, MetricCallable
from .pool import Pool
from .utils import resize_array


class Mosaic:
    def __init__(
        self,
        master: Master,
        pool: Pool,
        n_appearances: int = 1,
    ) -> None:
        """Construct a regular grid mosaic.

        Note:
            The Pool's tiles should all be the same size.

        Args:
            master: Master image to reconstruct.
            pool: Tile image pool with which to reconstruct the Master image.
            n_appearances: Number of times a tile can appear in the mosaic.

        Examples:
            Building a mosaic.

            >>> Mosaic(master, pool, n_appearances=1).build()
        """
        self._log = logging.getLogger(__name__)
        self.master = master
        if len(set([array.size for array in pool.arrays])) != 1:
            raise ValueError("Pool tiles sizes are not identical.")
        self.pool = pool
        self.tile_shape = (self.pool.arrays[0].shape[0], self.pool.arrays[0].shape[1])
        self.n_appearances = n_appearances
        self.grid = Grid(self.master, (self.size[1], self.size[0]), self.tile_shape)

    @property
    def size(self) -> Tuple[int, int]:
        """The size of the mosaic image.

        It can be different from the master image size as an integer number of
        tiles should fit within it.

        Returns:
            The width and height of the mosaic image.
        """
        return (
            self.master.array.shape[1]
            - self.master.array.shape[1] % self.tile_shape[1],
            self.master.array.shape[0]
            - self.master.array.shape[0] % self.tile_shape[0],
        )

    @property
    def n_leftover(self) -> int:
        """The number of tiles which will be unused when building the mosaic."""
        return len(self.pool) * self.n_appearances - len(self.grid.slices)

    def _d_matrix_worker(
        self, array: np.ndarray, metric_func: MetricCallable, **kwargs
    ) -> np.ndarray:
        """Parallel worker. Computes one row of the distance matrix."""
        # if the tile grid was subdivided the master array can be smaller
        # than the tiles, need to resize to match the shapes
        if array.shape[:-1] != self.tile_shape:
            # this isn't exact because we are upscalling the master array
            # we should be shrinking all the tile arrays but that is slower
            array = resize_array(array, (self.tile_shape[1], self.tile_shape[0]))
        return metric_func(array, self.pool.arrays, **kwargs)

    def compute_d_matrix(
        self,
        workers: int = 1,
        metric: Union[str, MetricCallable] = "norm",
        **kwargs,
    ) -> np.ndarray:
        """Compute the distance matrix between all the master's tiles and the
        pool tiles.

        Args:
            workers: The number of worker to use.
            metric: The distance metric used for the distance matrix. Either
                provide a string, for implemented metrics see ``phomo.metrics.METRICS``.
                Or a callable, which should take two ``np.ndarray``s and return a float.
            **kwargs: Passed to `metric`.

        Returns:
            Distance matrix, shape: (number of master arrays, number of tiles in the pool).
        """
        if isinstance(metric, str):
            if metric not in METRICS.keys():
                raise KeyError(
                    f"'%s' not in available metrics: %s",
                    metric,
                    repr(list(METRICS.keys())),
                )
            self._log.info("Using metric '%s'", metric)
            metric_func = METRICS[metric]
        else:
            self._log.info("Using user provided distance metric function.")
            metric_func = metric

        # Compute the distance matrix.
        worker = partial(self._d_matrix_worker, metric_func=metric_func, **kwargs)
        if workers != 1:
            self._log.info("Computing distance matrix with %i workers.", workers)
            with MpPool(processes=workers) as pool:
                d_matrix = np.array(
                    list(
                        tqdm(
                            pool.imap(
                                worker,
                                self.grid.arrays,
                                chunksize=len(self.grid) // workers,
                            ),
                            total=len(self.grid.slices),
                            desc="Building distance matrix",
                        )
                    )
                )
        else:
            # get rid of pool overhead if serial computation is desired.
            self._log.info("Computing distance matrix in serial.")
            d_matrix = np.array(
                [
                    worker(array)
                    for array in tqdm(self.grid.arrays, desc="Building distance matrix")
                ]
            )
        self._log.debug("d_matrix shape: %s", d_matrix.shape)
        return d_matrix

    def build(
        self,
        workers: int = 1,
        metric: Union[str, MetricCallable] = "norm",
        d_matrix: Optional[np.ndarray] = None,
        **kwargs,
    ) -> Image.Image:
        """Construct the mosaic image.

        Args:
            workers: The number of workers to use when computing the
                distance matrix.
            metric: The distance metric used for the distance matrix. Either
                provide a string, for implemented metrics see ``phomo.metrics.METRICS``.
                Or a callable, which should take two ``np.ndarray``s and return a float.
            d_matrix: Use a pre-computed distance matrix.
            **kwargs: Passed to the `metric` function.

        Returns:
            The PIL.Image instance of the mosaic.
        """
        mosaic = np.zeros((self.size[1], self.size[0], 3))

        # Compute the distance matrix.
        if d_matrix is None:
            d_matrix = self.compute_d_matrix(workers=workers, metric=metric, **kwargs)

        # Keep track of tiles and sub arrays.
        placed_master_arrays = set()
        placed_tiles = set()
        n_appearances = [0] * len(self.pool)

        pbar = tqdm(total=d_matrix.shape[0], desc="Building mosaic")
        # from: https://stackoverflow.com/questions/29046162/numpy-array-loss-of-dimension-when-masking
        sorted_master_slices_i, sorted_tiles = np.unravel_index(
            np.argsort(d_matrix, axis=None), d_matrix.shape
        )
        for slices_i, tile in zip(sorted_master_slices_i, sorted_tiles):
            if slices_i in placed_master_arrays or tile in placed_tiles:
                continue
            slices = self.grid.slices[slices_i]
            tile_array = self.pool.arrays[tile]
            # if the grid has been subdivided then the tile should be shrunk to
            # the size of the subdivision
            array_size = (
                slices[1].stop - slices[1].start,
                slices[0].stop - slices[0].start,
            )
            if tile_array.shape[:-1] != array_size[::-1]:
                tile_array = resize_array(tile_array, array_size)

            # shift slices back so that the centering of the mosaic within the
            # master image is removed
            slices = self.grid.remove_origin(slices)
            mosaic[slices[0], slices[1]] = tile_array
            placed_master_arrays.add(slices_i)
            n_appearances[tile] += 1
            if n_appearances[tile] == self.n_appearances:
                placed_tiles.add(tile)
            pbar.update(1)
        pbar.close()
        return Image.fromarray(np.uint8(mosaic))

    def __repr__(self) -> str:
        # indent these guys
        master = repr(self.master).replace("\n", "\n    ")
        pool = repr(self.pool).replace("\n", "\n    ")
        grid = repr(self.grid).replace("\n", "\n    ")
        return f"""{self.__class__.__module__}.{self.__class__.__name__} at {hex(id(self))}:
    n_appearances: {self.n_appearances}
    mosaic size: {self.size}
    tile shape: {self.tile_shape}
    leftover tiles: {self.n_leftover}
    {grid}
    {master}
    {pool}"""



import logging
from pathlib import Path
from typing import Optional, Tuple

import numpy as np
from PIL import Image

from .palette import Palette
from .utils import open_img_file


class Master(Palette):
    @classmethod
    def from_file(
        cls,
        master_image_file: Path,
        *args,
        crop_ratio: Optional[float] = None,
        img_size: Optional[Tuple[int, int]] = None,
        convert: Optional[str] = None,
        **kwargs,
    ) -> "Master":
        """Create a master image from file.

        Args:
            master_image_file: path to image file.
            crop_ratio: width to height ratio to crop the master image to. 1 results in a square image.
            image_size: resize the image to the provided size, width followed by height.
            convert: convert the image to the provided mode. See PIL Modes.

        Returns:
            Master instance.

        Examples:
            For black and white square 1280x1280 image.

            >>> Master.from_file("master.png", crop_ratio=1, img_size=(1280, 1280), convert="L")
        """
        img = open_img_file(
            master_image_file, crop_ratio=crop_ratio, size=img_size, convert=convert
        )
        return cls.from_image(img, *args, **kwargs)

    @classmethod
    def from_image(cls, master_image: Image.Image, *args, **kwargs) -> "Master":
        """Create a master image from PIL.Image.Image

        Args:
            master_image: PIL.Image instance.

        Returns:
            Master instance.
        """
        array = np.asarray(master_image)
        # make sure the arrays have 3 channels even in black and white
        if array.ndim == 2:
            array = np.stack([array] * 3, -1)
        return cls(array, *args, **kwargs)

    def __init__(self, array: np.ndarray) -> None:
        """The master image.

        Args:
            array: numpy array of the image, should contain 3 channels.

        Returns:
            Master instance.
        """
        self.array = array
        self._log = logging.getLogger(__name__)
        self._log.info("master shape: %s", self.array.shape)

    @property
    def img(self):
        """PIL.Image of the Master image."""
        return Image.fromarray(self.array.round(0).astype("uint8"), mode="RGB")

    @property
    def pixels(self) -> np.ndarray:
        """Array containing the 3-channel pixel values of the Master image."""
        return self.array.reshape(-1, self.array.shape[-1])

    def __repr__(self) -> str:
        return f"""{self.__class__.__module__}.{self.__class__.__name__} at {hex(id(self))}:
    shape: {self.array.shape}"""



from pathlib import Path
from typing import Optional, Tuple

import numpy as np
from PIL import Image
from PIL.ImageOps import exif_transpose
from tqdm.auto import tqdm


def rainbow_of_squares(
    target_dir: Path, size: Tuple[int, int] = (10, 10), range_params=(0, 256, 15)
) -> None:
    """Generate 5832 small solid-color tiles for experimentation and testing.

    Args:
        target_dir: direcotry in which to place the rainbow tiles.
        size: size of the images, width followed by height.
        range_params: Passed to ``range()`` to stride through each color channel.
    """
    target_dir.mkdir(exist_ok=True)
    with tqdm(total=3 * len(range(*range_params))) as pbar:
        canvas = np.ones((*size[::-1], 3))
        for r in range(*range_params):
            for g in range(*range_params):
                for b in range(*range_params):
                    img = (canvas * [r, g, b]).astype(np.uint8)
                    filename = "{:03d}-{:03d}-{:03d}.png".format(r, g, b)
                    img = Image.fromarray(img, mode="RGB")
                    img.save(target_dir / filename)
                    pbar.update()


def crop_to_ratio(image: Image.Image, ratio: float = 1) -> Image.Image:
    """Reshapes an image to the specified ratio by cropping along the larger
    dimension.

    Args:
        image: PIL.Image to crop.
        ratio: width to height ratio to which to crop the image. Use 1 to obtain a square image.

    Returns:
        Cropped PIL.Image.
    """

    width, height = image.size

    def crop_height(image, rx):
        return image.crop(
            (
                0,
                (rx / 2),
                width,
                height - (rx / 2),
            )
        )

    def crop_width(image, rx):
        return image.crop(
            (
                (rx / 2),
                0,
                width - (rx / 2),
                height,
            )
        )

    # Find the delta change.
    rxheight = width / ratio - height
    rxwidth = height * ratio - width

    # Can only crop pixels, not add them.
    if rxheight < 0 and rxwidth < 0:
        # If both sides can be cropped to get what we want:
        # Select the largest (because both are negative)
        if rxheight > rxwidth:
            return crop_height(image, rxheight * -1)
        else:
            return crop_width(image, rxwidth * -1)

    elif rxheight < 0:
        # Trim height to fit aspect ratio
        return crop_height(image, rxheight * -1)

    elif rxwidth < 0:
        # Trim width to fit aspect ratio
        return crop_width(image, rxwidth * -1)

    else:
        # Can't do anything in this case
        return image


def open_img_file(
    img_file: Path,
    crop_ratio: Optional[float] = None,
    size: Optional[Tuple[int, int]] = None,
    convert: Optional[str] = None,
) -> Image.Image:
    """Open an image file with some extra bells and whistles.

    Args:
        img_file: path to the image.
        crop_ratio: width to height to which to crop the image.
        size: resize image.
        convert: convert the image to the provided mode. See PIL image modes.

    Returns:
        Image instance.
    """
    with Image.open(img_file) as img:
        img = exif_transpose(img)
        if crop_ratio is not None:
            img = crop_to_ratio(img, crop_ratio)
        if size is not None:
            img = img.resize(size)
        if convert is not None:
            img = img.convert(convert)
        else:
            img = img.convert("RGB")
    return img


def resize_array(
    array: np.ndarray, size: Tuple[int, int], *args, **kwargs
) -> np.ndarray:
    """Resize an array representing and image.

    Args:
        array: array containing the images data.
        size: desired size, width followed by height.
        *args, **kwargs: passed to `PIL.Image.resize`.

    Returns:
        Array containing the resized image data.
    """
    return np.asarray(Image.fromarray(array).resize(size, *args, **kwargs))



import sys

# prior to python 3.8, Protocol is in typing_extensions
if sys.version_info[0] == 3 and sys.version_info[1] < 8:
    from typing_extensions import Protocol
else:
    from typing import Protocol

import numpy as np

METRICS = {}


class MetricCallable(Protocol):  # type: ignore
    def __call__(self, master_chunk: np.ndarray, tile_arrays: np.ndarray) -> np.ndarray:
        ...


def register_metric(func):
    METRICS[func.__name__] = func
    return func


@register_metric
def greyscale(
    master_chunk: np.ndarray, tile_arrays: np.ndarray, **kwargs
) -> np.ndarray:
    """Compute the greyscale distance.

    This metric ignores colours and compares greyscale values. Should provide better
    photomosaics when using few tiles images.


    Args:
        master_chunk: array containing the RGB pixels with values between 0 and 255.
        tile_arrays: array tile pixel arrays, values between 0 and 255.
        **kwargs: passed to ``np.linalg.norm``.

    Returns:
        Colour distance approximation between the master chunk and all the tiles
            arrays.
    """
    delta = np.subtract(
        master_chunk.sum(axis=-1), tile_arrays.sum(axis=-1), dtype=float
    )
    return np.linalg.norm(delta.reshape(delta.shape[0], -1), axis=-1, **kwargs)


@register_metric
def norm(master_chunk: np.ndarray, tile_arrays: np.ndarray, **kwargs) -> np.ndarray:
    """``np.linalg.norm`` distance metric.

    Quick distance metric in RGB space.

    Args:
        master_chunk: array containing the RGB pixels with values between 0 and 255.
        tile_arrays: list of tile pixel arrays, values between 0 and 255.
        **kwargs: passed to ``np.linalg.norm``.

    Returns:
        Colour distance approximation between the master chunk and all the tiles
            arrays.
    """
    return np.linalg.norm(
        np.subtract(master_chunk, tile_arrays, dtype=float).reshape(
            tile_arrays.shape[0], -1, tile_arrays.shape[-1]
        ),
        axis=(1, 2),
        **kwargs,
    )


# def norm(master_chunk: np.ndarray, tile_arrays: np.ndarray, **kwargs) -> float:
#     """`np.linalg.norm` distance metric.

#     Args:
#         master_chunk: array containing the RGB pixels with values between 0
#             and 255.
#         tile_arrays: array containing the RGB pixels with values between 0
#             and 255.

#     Returns:
#         Colour distance approximation.
#     """
#     return np.linalg.norm(
#         np.linalg.norm(
#             np.subtract(master_chunk, tile_arrays, dtype=float),
#             axis=-1,
#
#             **kwargs,
#         )
#     )


@register_metric
def luv_approx(
    master_chunk: np.ndarray, tile_arrays: np.ndarray, **kwargs
) -> np.ndarray:
    """Distance metric using a L*U*V space approximation.

    This metric should provide more accurate colour matching.

    Reference:
        https://www.compuphase.com/cmetric.htm

    Args:
        master_chunk: array containing the RGB pixels with values between and 255.
        tile_arrays: array containing the RGB pixels with values between 0 and 255.
        **kwargs: passed to ``np.linalg.norm``.

    Returns:
        Colour distance approximation between the master chunk and all the tiles
            arrays.
    """
    r = (master_chunk[:, :, 0] + tile_arrays[:, :, :, 0]) // 2
    d = np.subtract(master_chunk, tile_arrays, dtype=float)
    return np.linalg.norm(
        (
            ((512 + r) * d[:, :, :, 0] ** 2)
            + 1024 * d[:, :, :, 1] ** 2
            + ((767 - r) * d[:, :, :, 2] ** 2)
        ).reshape(d.shape[0], -1),
        axis=-1,
        **kwargs,
    )



import argparse
import logging
import sys
from pathlib import Path
from typing import List

from . import logger
from .metrics import METRICS
from .mosaic import Master, Mosaic, Pool


def parse_args(args: List[str]) -> argparse.Namespace:
    """Parse the command line arguments.

    Args:
        args: list of command line arguments.

    Returns:
        argparse Namespace of the parsed arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "master",
        help="Master image path.",
        type=str,
    )
    parser.add_argument(
        "tile_dir",
        help="Directory containing the tile images.",
        type=str,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="Mosiac output path.",
        type=str,
        default=None,
    )
    parser.add_argument(
        "-c",
        "--master-crop-ratio",
        help="Crop the master image to width/height ratio.",
        default=None,
        type=float,
    )
    parser.add_argument(
        "-s",
        "--master-size",
        help="Resize master image to width, height.",
        type=int,
        nargs="+",
        default=None,
    )

    parser.add_argument(
        "-C",
        "--tile-crop-ratio",
        help="Crop the tile images to width/height ratio.",
        default=None,
        type=float,
    )
    parser.add_argument(
        "-S",
        "--tile-size",
        help="Resize tile images to width, height.",
        type=int,
        nargs="+",
        default=None,
    )
    parser.add_argument(
        "-n",
        "--n-appearances",
        help="The number of times a tile can appear in the mosaic.",
        type=int,
        default=1,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Verbosity.",
        action="count",
        default=0,
    ),
    parser.add_argument(
        "-b",
        "--black-and-white",
        help="Convert master and tile images to black and white.",
        action="store_true",
    )
    parser.add_argument(
        "-g",
        "--show-grid",
        help="Show the tile grid, don't build the mosiac.",
        action="store_true",
    )
    parser.add_argument(
        "-d",
        "--subdivisions",
        help="Grid subdivision thresholds.",
        nargs="+",
        default=[],
        type=float,
    )
    parser.add_argument(
        "-m",
        "--metric",
        help="Distance metric.",
        default="norm",
        type=str,
        choices=METRICS.keys(),
    )
    parser.add_argument(
        "-j",
        "--workers",
        help="Number of workers use to run when computing the distance matrix.",
        default=1,
        type=int,
    )
    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])

    if args.verbose > 0:
        verbose_map = {1: logging.INFO, 2: logging.DEBUG}
        level = verbose_map[args.verbose]
        # from https://docs.python.org/3/howto/logging.html#configuring-logging
        logger.setLevel(level)
        # create console handler and set level to debug
        handler = logging.StreamHandler()
        handler.setLevel(level)
        # create formatter
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        # add formatter to handler
        handler.setFormatter(formatter)
        # add ch to logger
        logger.addHandler(handler)

    mode = None
    if args.black_and_white:
        mode = "L"

    tile_size = args.tile_size
    if args.tile_size is not None:
        if len(args.tile_size) != 2:
            raise ValueError("Provided tile size is not of length 2.")
        tile_size = tuple(tile_size)

    master_size = args.master_size
    if args.master_size is not None:
        if len(args.master_size) != 2:
            raise ValueError("Provided master size is not of length 2.")
        master_size = tuple(master_size)

    master = Master.from_file(
        Path(args.master),
        crop_ratio=args.master_crop_ratio,
        img_size=master_size,  # type: ignore
        convert=mode,
    )

    pool = Pool.from_dir(
        Path(args.tile_dir),
        crop_ratio=args.tile_crop_ratio,
        tile_size=tile_size,  # type: ignore
        convert=mode,
    )

    mosaic = Mosaic(master, pool, n_appearances=args.n_appearances)
    for threshold in args.subdivisions:
        mosaic.grid.subdivide(threshold)

    logger.debug("mosaic:\n%s", repr(mosaic))
    logger.info("Number of unused tiles: %i", mosaic.n_leftover)

    if args.show_grid:
        grid_im = mosaic.grid.plot()
        grid_im.show()
    else:
        mosaic_im = mosaic.build(workers=args.workers, metric=args.metric)
        if args.output is None:
            mosaic_im.show()
        else:
            mosaic_im.save(args.output)
